/**
 * Classe Renderer, junto com a classe ExemploJava, mostra um exemplo de 
 * como trabalhar com Java e OpenGL utilizando a API JOGL.  aberta uma 
 * janela na qual  desenhada uma casinha.  possvel fazer 
 * zoom in e zoom out usando o mouse, e mover a posio do observador 
 * virtual com as teclas de setas, HOME e END.
 * 
 * @author Marcelo Cohen
 * @version 1.0
 */

import java.awt.event.*; 
import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import com.sun.opengl.util.*;

import OBJLoader.*;

public class Renderer extends MouseAdapter implements GLEventListener, KeyListener
{
	// Atributos
	private GL gl;
	private GLU glu;
	private GLUT glut;
	private GLAutoDrawable glDrawable;
	private double angle, fAspect;
	private float rotX, rotY, obsZ;
  	private OBJModel model;
	private String name;
	private float maxSize;

  	private static final float MAX_SIZE = 2.0f;  // for a model's dimension

	/**
	 * Construtor da classe Renderer que no recebe parmetros.
	 */
	public Renderer(String name)
	{
		// Especifica o ngulo da projeo perspectiva  
		angle=60;
		// Posio do observador virtual
		rotX = 5;
		rotY = 0;
		obsZ = 2; 
		// Inicializa o valor para correo de aspecto   
		fAspect = 1;    
		this.name = name;
		this.maxSize = MAX_SIZE;
	}

	/**
	 * Mtodo definido na interface GLEventListener e chamado pelo objeto no qual ser feito o desenho
	 * logo aps a inicializao do contexto OpenGL. 
	 */    
	public void init(GLAutoDrawable drawable)
	{
		glDrawable = drawable;
		gl = drawable.getGL();
		// glu = drawable.getGLU();       
		glu = new GLU();
		glut = new GLUT();

		drawable.setGL(new DebugGL(gl));        

		// Carrega o modelo
		model = new OBJModel(name, maxSize, gl, true);

		gl.glClearColor(0.0f, 0.0f, 0.7f, 1.0f);

		// Habilita GL_COLOR_MATERIAL, que significa alterar a cor do material
		// quando a cor corrente  alterada via glColor...
		//gl.glEnable(GL.GL_COLOR_MATERIAL);
		gl.glColorMaterial(GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT_AND_DIFFUSE);

		// Habilita iluminao
		gl.glEnable(GL.GL_LIGHTING);

		// Habilita luz 0
		gl.glEnable(GL.GL_LIGHT0);

		// Habilita Z-Buffer
		gl.glEnable(GL.GL_DEPTH_TEST);

		// Habilita tonalizao Gouraud
		gl.glShadeModel(GL.GL_SMOOTH);         
    
		gl.glTexEnvi(GL.GL_TEXTURE_ENV, GL.GL_TEXTURE_ENV_MODE, GL.GL_MODULATE);

		// Especifica sistema de coordenadas de projeo
		gl.glMatrixMode(GL.GL_PROJECTION);
		// Inicializa sistema de coordenadas de projeo
		gl.glLoadIdentity();
		// Especifica a projeo perspectiva(angulo,aspecto,zMin,zMax)
		glu.gluPerspective(angle, fAspect, 0.5, 500);
		// Volta para Modelview
		gl.glMatrixMode(GL.GL_MODELVIEW);

	}

	/**
         * Mtodo que desenha a casinha
         */
        public void desenhaCasa()
	{
		gl.glBegin(GL.GL_QUADS);
		// Frente
		gl.glColor3f( 0.3f, 0.5f, 0.4f );
		gl.glNormal3f(  0f,  0f,  1 );			// normal aponta para fora da tela
		gl.glVertex3f( -1,  0,  1 );
		gl.glVertex3f( -1,  2,  1 );
		gl.glVertex3f(  1,  2,  1 );
		gl.glVertex3f(  1,  0,  1 );

		// Fundos
		gl.glNormal3f(  0,  0, -1 );			// normal aponta para dentro da tela
		gl.glVertex3f( -1,  0, -1 );
		gl.glVertex3f(  1,  0, -1 );
		gl.glVertex3f(  1,  2, -1 );
		gl.glVertex3f( -1,  2, -1 );
			
		// Lado esquerdo
		gl.glNormal3f( -1,  0,  0 );			// normal aponta para a esquerda
	  	gl.glVertex3f( -1,  0,  1 );
	  	gl.glVertex3f( -1,  2,  1 );
	  	gl.glVertex3f( -1,  2, -1 );
	  	gl.glVertex3f( -1,  0, -1 );

		// Lado direito
		gl.glNormal3f(  1,  0,  0 );			// normal aponta para a direita
		gl.glVertex3f(  1,  0,  1 );
		gl.glVertex3f(  1,  0, -1 );
		gl.glVertex3f(  1,  2, -1 );
		gl.glVertex3f(  1,  2,  1 );

		// Finaliza os 4 quads
		gl.glEnd();

		// Teto  uma pirmide de base quadrangular
		gl.glBegin(GL.GL_TRIANGLES);
		
		// Frente
		gl.glColor3f( 0.8f, 0.2f, 0);

		// As normais do teto no so unitrias (ex: 0 1 1), portanto eu apliquei manualmente o clculo
		// de normalizao. Se voc no fizer isso, a iluminao ficar INCORRETA.
	  
		gl.glNormal3f(  0,  0.7f,  0.7f );		// normal diagonal para cima e para fora da tela
	  	gl.glVertex3f( -1,  2,  1 );
	  	gl.glVertex3f(  0,  4,  0 );
	  	gl.glVertex3f(  1,  2,  1 );

		// Lado direito
	  	gl.glNormal3f(  0.7f,  0.7f,  0 );		// normal diagonal para cima e para a direita
	  	gl.glVertex3f(  1,  2,  1 );
	  	gl.glVertex3f(  1,  2, -1 );
	  	gl.glVertex3f(  0,  4,  0 );

		// Traseira
		gl.glNormal3f(  0,  0.7f, -0.7f );		// normal diagonal para cima up e para dentro da tela
	  	gl.glVertex3f(  1,  2, -1 );
	  	gl.glVertex3f( -1,  2, -1 );
	  	gl.glVertex3f(  0,  4,  0 );

		// Lado esquerdo
	  	gl.glNormal3f( -0.7f,  0.7f,  0 );		// normal diagonal para cima e para a esquerda
	  	gl.glVertex3f( -1,  2,  1 );
	  	gl.glVertex3f(  0,  4,  0 );
	  	gl.glVertex3f( -1,  2, -1 );

		// Finaliza os 4 tringulos
		gl.glEnd();
	}

	/**
         * Mtodo que desenha o cho
         */
        public void desenhaChao()
	{
		// Cho embaixo da casinha
		gl.glBegin(GL.GL_QUADS);
	  	gl.glColor3f( 0.2f, 0.7f, 0.3f );
	  	gl.glNormal3f(  0,  1,  0 );	// normal aponta diretamente para cima
	  	gl.glVertex3f(-10,  0, -10);
	  	gl.glVertex3f( 10,  0, -10);
	  	gl.glVertex3f( 10,  0,  10);
	  	gl.glVertex3f(-10,  0,  10);
		gl.glEnd();
	}

	/**
         * Mtodo que desenha um cho melhorado (quadriculado)
         */
        public void desenhaChaoGrade()
	{
		gl.glBegin(GL.GL_QUADS);
		float z,x;
		float tam = 2.0f;	// define o tamanho de cada quadrado
		boolean passo = false;
		for(z=-30;z<=30;z+=tam)
		{
			for(x=-30;x<=30;x+=tam)
			{
				// Alterna entre as duas cores
				if(passo) gl.glColor3f( 0.2f, 0.2f, 0.5f );
				else gl.glColor3f( 0.7f, 0.7f, 0.7f );
				passo = !passo;

			  	gl.glNormal3f(  0,  1,  0 );	// normal aponta diretamente para cima
			  	gl.glVertex3f(x,  0, z);
			  	gl.glVertex3f(x+tam,  0, z);
			  	gl.glVertex3f(x+tam,  0, z+tam);
			  	gl.glVertex3f(x,  0,  z+tam);
			}
		}
		gl.glEnd();
	}

	/**
	 * Mtodo definido na interface GLEventListener e chamado pelo objeto no qual ser feito o desenho
	 * para comear a fazer o desenho OpenGL pelo cliente.
	 */  
	public void display(GLAutoDrawable drawable)
	{
		gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT );
		gl.glLoadIdentity();    

		//especificaParametrosVisualizacao(); 

		defineIluminacao();
		posicionaObservador();

		gl.glDisable(GL.GL_COLOR_MATERIAL);
    		model.drawList(gl);      // draw the model

		// Desenha uma casinha
		//desenhaCasa();

		// Desenha um cho
		//desenhaChao();
		//gl.glDisable(GL.GL_TEXTURE_2D);
		gl.glEnable(GL.GL_COLOR_MATERIAL);
		//desenhaChaoGrade();
	}

	/**
	 * Mtodo definido na interface GLEventListener e chamado pelo objeto no qual ser feito o desenho
	 * depois que a janela foi redimensionada.
	 */  
	public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height)
	{
		gl.glViewport(0, 0, width, height);
		fAspect = (float)width/(float)height;      
		especificaParametrosVisualizacao();
	}

	/**
	 * Mtodo definido na interface GLEventListener e chamado pelo objeto no qual ser feito o desenho
	 * quando o modo de exibio ou o dispositivo de exibio associado foi alterado.
	 */  
	public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) { }

	/**
	 * Mtodo responsvel pela especificao dos parmetros de iluminao.
	 */     
	public void defineIluminacao ()
	{    
		float luzAmbiente[]={0.3f, 0.3f, 0.3f, 1.0f}; 
		float luzDifusa[]={0.7f, 0.7f, 0.7f, 1.0f};  
		float luzEspecular[]={0.8f, 0.8f, 0.8f, 1.0f};
		float posicaoLuz[]={100.0f, 200.0f, 100.0f, 1.0f};  
		float direcaoLuz[]={-10.0f, -20.0f, -10.0f, 1.0f};

		float anguloSpot = 60; // o quanto o spot ilumina, em graus a partir do centro
		float expoenteSpot = 120; // entre 0 e 128, define a distribuio de luz a partir do centro do spot
		// 0 = uniforme
		// 1 - focado

		// Capacidade de brilho do material
		float especularidade[]={1.0f, 1.0f, 1.0f, 1.0f};
		int especMaterial = 60;

		// Define a reflectncia do material 
		gl.glMaterialfv(GL.GL_FRONT, GL.GL_SPECULAR, especularidade, 0);
		// Define a concentrao do brilho
		gl.glMateriali(GL.GL_FRONT, GL.GL_SHININESS, especMaterial);

		// Ativa o uso da luz ambiente 
		gl.glLightModelfv(GL.GL_LIGHT_MODEL_AMBIENT, luzAmbiente, 0);

		// Define os parmetros da luz de nmero 0
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_AMBIENT, luzAmbiente, 0); 
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, luzDifusa, 0 );
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, luzEspecular, 0);
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, posicaoLuz, 0 );   
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_SPOT_DIRECTION, direcaoLuz, 0 ); 

		// Comente a linha a seguir para DESATIVAR o spot
		gl.glLightf(GL.GL_LIGHT0, GL.GL_SPOT_CUTOFF, anguloSpot); 

		gl.glLightf(GL.GL_LIGHT0, GL.GL_SPOT_EXPONENT, expoenteSpot); 
	}

	/**
	 * Mtodo usado para especificar a posio do observador virtual.
	 */    
	public void posicionaObservador()
	{
		// Especifica sistema de coordenadas do modelo
		gl.glMatrixMode(GL.GL_MODELVIEW);
		// Inicializa sistema de coordenadas do modelo  
		gl.glLoadIdentity();
		// Especifica posio do observador e do alvo   
		gl.glTranslatef(0.0f, 0.0f, -obsZ);    
		gl.glRotatef(rotX, 1.0f, 0.0f, 0.0f);  
		gl.glRotatef(rotY, 0.0f, 1.0f, 0.0f);
		//defineIluminacao();
	}
   
	/**
	 * Mtodo usado para especificar o volume de visualizao.
	 */    
	public void especificaParametrosVisualizacao()
	{
		// Especifica sistema de coordenadas de projeo
		gl.glMatrixMode(GL.GL_PROJECTION);
		// Inicializa sistema de coordenadas de projeo
		gl.glLoadIdentity();
		// Especifica a projeo perspectiva(angulo,aspecto,zMin,zMax)
		glu.gluPerspective(angle, fAspect, 0.5, 500);
		gl.glMatrixMode(GL.GL_MODELVIEW);

		//posicionaObservador();
	}

	/**
	 * Mtodo da classe MouseAdapter que est sendo sobrescrito para gerenciar os 
	 * eventos de clique de mouse, de maneira que sej feito zoom in e zoom out.
	 */  
	public void mouseClicked(MouseEvent e)
	{
		if (e.getButton() == MouseEvent.BUTTON1) // Zoom in
		if (angle >= 4) angle -= 4;
		if (e.getButton() == MouseEvent.BUTTON3) // Zoom out
		if (angle <= 72) angle += 4;
		glDrawable.display();
	}

	/**
	 * Mtodo definido na interface KeyListener que est sendo implementado para, 
	 * de acordo com as teclas pressionadas, permitir mover a posio do observador
	 * virtual.
	 */        
	public void keyPressed(KeyEvent e)
	{
		switch (e.getKeyCode())
		{
			case KeyEvent.VK_RIGHT:		rotY++;   
										break;
			case KeyEvent.VK_LEFT:		rotY--;   
										break;     
			case KeyEvent.VK_UP:		rotX++;   
										break;                              
			case KeyEvent.VK_DOWN:		rotX--;   
										break;                               
			case KeyEvent.VK_HOME:		obsZ++;   
										break;                                  
			case KeyEvent.VK_END:		obsZ--;   
										break;  
			case KeyEvent.VK_ESCAPE:	System.exit(0);
										break;
		}  
		glDrawable.display();
	}

	/**
	 * Mtodo definido na interface KeyListener.
	 */      
	public void keyTyped(KeyEvent e) { }

	/**
	 * Mtodo definido na interface KeyListener.
	 */       
	public void keyReleased(KeyEvent e) { }
}

